-------------------------------------------------------------------------------
--! @file       HANUMAN.vhd
--! @author     Ahmed Ferozpuri
--! @brief      Datapath for PRIMATEs-HANUMAN
-------------------------------------------------------------------------------

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE work.PRIMATEsPKG.ALL;

ENTITY HANUMAN IS
	GENERIC (  
			S : INTEGER := 80;
			V : INTEGER := 200
	);
	PORT	(
			--GMU INTERFACE SIGNALS
			clk                       : IN  std_logic;
			key                       : IN  std_logic_vector(S		    -1 DOWNTO 0);
			bdi                       : IN  std_logic_vector(RATE       -1 DOWNTO 0);
			key_ready                 : IN  std_logic;
			bdi_pad_loc 			  : IN std_logic_vector	(RATE/8		-1 DOWNTO 0);
			bdo                       : OUT std_logic_vector(RATE       -1 DOWNTO 0);
			bdo_sel			          : IN std_logic;
			msg_auth_valid            : OUT std_logic;
		  
			--USER DEFINED SIGNALS (FROM CIPHERCORE)
			init                      : IN  std_logic;
			run                       : IN  std_logic;
			mode                      : IN  std_logic;
			lastBlockFull             : IN  std_logic;
			selV                      : IN  std_logic_vector(2 			-1 DOWNTO 0);
			pn 						  : IN  std_logic_vector(2 			-1 DOWNTO 0);
			done	                  : OUT std_logic;
		  
			--SIPO/PISO SIGNALS
			npub_en                   : IN  std_logic;
			etag_en                   : IN  std_logic;
			tag_en                    : IN  std_logic;
			tag_ld                    : IN  std_logic
	);
END HANUMAN;

ARCHITECTURE behavioral OF HANUMAN IS

	SIGNAL ViPN, ViPNVal, Mi0, Mi1, Mi2, V_ol:std_logic_vector(V-1 DOWNTO 0);
	SIGNAL KxorVC2, key_r, npub, tag, exp_tag : std_logic_vector(S-1 DOWNTO 0);
	SIGNAL R0, XNxorVR, tag_word : std_logic_vector(RATE-1 DOWNTO 0);
	SIGNAL pad_seq : std_logic_vector(RATE/8-1 DOWNTO 0);
	
BEGIN
	
	R0 <= (OTHERS => '0');
	Mi0 <= R0 & key_r & npub;
	Mi1 <= XNxorVR & V_ol(V-RATE-1 DOWNTO 0);
	Mi2(V-RATE-1 DOWNTO 0) <= V_ol(V-RATE-1 DOWNTO 0);
	ViPN <= ViPNVal(V-1 DOWNTO V-RATE) & (ViPNVal(V-RATE-1) XOR lastBlockFull) & ViPNVal(V-RATE-2 DOWNTO 0);
	
	bdo <= XNxorVR when bdo_sel = '1' else tag_word;
	msg_auth_valid <= '1' WHEN exp_tag = KxorVC2 ELSE '0';
	tag <= KxorVC2;

	WITH selV SELECT
	ViPNVal <= 	Mi0 WHEN "01",
				Mi1 WHEN "00",
				Mi2 WHEN "10",
				V_ol WHEN OTHERS;
	
	pad_seq <= 	"00001" WHEN bdi_pad_loc(0) = '1' ELSE
				"00011" WHEN bdi_pad_loc(1) = '1' ELSE
				"00111" WHEN bdi_pad_loc(2) = '1' ELSE
				"01111" WHEN bdi_pad_loc(3) = '1' ELSE
				"00000";

	GEN_S: FOR i IN 0 TO S-1 GENERATE
		KxorVC2(S-1-i) <= key_r(S-1-i) XOR V_ol(V-RATE-1-i);
	END GENERATE GEN_S;
				  
	GEN_R: FOR i IN 0 TO RATE-1 GENERATE
		XNxorVR(RATE-1-i) <= bdi(RATE-1-i) XOR V_ol(V-1-i);
		Mi2(V-1-i) <= bdi(RATE-1-i) XOR V_ol(V-1-i) WHEN pad_seq(RATE/8-1-i/8) = '1' ELSE bdi(RATE-1-i);
	END GENERATE GEN_R;

	PN_SM: ENTITY work.PN(behavioral)
	GENERIC MAP (V=>V, isINV=>0)
	PORT MAP (clk => clk, init => init, run => run, mode => mode, pn => pn, V_i => ViPN, V_o => V_ol, done => done);
	
	REGK: ENTITY work.REGN_aqf(behavioral)
	GENERIC MAP (N => S)
	PORT MAP(	clk => clk,	en => key_ready, D => key,	Q => key_r);
	
	SIPO_NPUB: ENTITY work.SIPON_M(behavioral)
	GENERIC MAP (N=>RATE, M=>S)
	PORT MAP (clk => clk, en => npub_en, D => bdi, Q => npub);

	SIPO_ETAG: ENTITY work.SIPON_M(behavioral)
	GENERIC MAP (N=>RATE, M=>S)
	PORT MAP (clk => clk, en => etag_en, D => bdi, Q => exp_tag);

	PISO_TAG: ENTITY work.PISON_M(behavioral)
	GENERIC MAP (N=>S, M=>RATE)
	PORT MAP (clk => clk, ld => tag_ld, en => tag_en, D => tag, Q => tag_word);
	
END behavioral;