-------------------------------------------------------------------------------
--! @file       PCCTRL.vhd
--! @author     Ahmed Ferozpuri
--! @brief      PRIMATEs-HANUMAN Cipher Controller
-------------------------------------------------------------------------------
LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
USE work.PRIMATEsPKG.ALL;

ENTITY PCCTRL IS
	PORT(	
		  --GMU INTERFACE SIGNALS
        clk                       : IN  std_logic;
        rst                       : IN  std_logic;

        key_ready                 : OUT  std_logic;
		key_valid                 : IN  std_logic;
        key_update                : IN std_logic;
		bdi_ready                 : OUT  std_logic;
		bdi_valid                 : IN  std_logic;
		bdi_type                  : IN  std_logic_vector(3                -1 downto 0);
		
        decrypt                   : IN  std_logic;
        bdi_eot                   : IN  std_logic;
        bdi_eoi                   : IN  std_logic;
        bdi_size                  : IN  std_logic_vector(3              			    downto 0);

        bdo_sel			          : OUT std_logic;
		bdo_ready                 : IN  std_logic;
        bdo_valid                 : OUT std_logic;
        msg_auth_done             : OUT std_logic;
		  
		--USER DEFINED SIGNALS (TO DATAPATH)
		init                      : OUT  std_logic;
		run                       : OUT  std_logic;
		mode                      : OUT  std_logic;
		lastBlockFull             : OUT  std_logic;
		selV                      : OUT  std_logic_vector(1 								 DOWNTO 0);
		pn 						  : OUT  std_logic_vector(1 								 DOWNTO 0);
		done	                  : IN std_logic;
		
		--SIPO/PISO SIGNALS
		npub_en                   : OUT  std_logic;
		etag_en                   : OUT  std_logic;
		tag_en                    : OUT  std_logic;
		tag_ld                    : OUT  std_logic
		);
END PCCTRL;

ARCHITECTURE behavioral OF PCCTRL IS
	TYPE hStates IS (	IDLE, ACTIVATE_KEY, LOAD_NPUB, LOAD_INIT, RUN_INIT,
						LOAD_DATA, RUN_BLOCK,
						GENVER_TAG, WRITE_TAG);
	SIGNAL hps, hns : hStates;
	
	SIGNAL bdi_eot_r, bdi_eot_l, bdi_eoi_r, bdi_eoi_l, bdi_type_l, decrypt_l, decrypt_r : std_logic;
	SIGNAL blockFull, blockEmpty : std_logic;
	SIGNAL bdi_type_r : std_logic_vector(3			-1 downto 0);
	
BEGIN

	blockFull <= '1' WHEN bdi_size = "0101" ELSE '0';
	blockEmpty <= '1' WHEN bdi_size = "0000" ELSE '0';
	lastBlockFull <= bdi_eot and blockFull;
	mode <= decrypt;
	
	HANUMAN_STATE: PROCESS (clk, rst)
	BEGIN
		IF rst = '1' THEN
			hps <= IDLE;
			bdi_eoi_r <= '0';
			bdi_eot_r <= '0';
			decrypt_r <= '0';
			bdi_type_r <= (others => '0');
		ELSIF rising_edge(clk) THEN
			hps <= hns;
			
			IF (bdi_eoi_l = '1') THEN 
				bdi_eoi_r <= bdi_eoi;
			END IF;
			
			IF (bdi_eot_l = '1') THEN 
				bdi_eot_r <= bdi_eot;
			END IF;
			
			IF (bdi_type_l = '1') THEN
				bdi_type_r <= bdi_type;
			END IF;
			
			IF (decrypt_l = '1') THEN
				decrypt_r <= decrypt;
			END IF;
		END IF;
	END PROCESS HANUMAN_STATE;
	
	HANUMAN_OUTPUT: PROCESS (rst, hps, key_valid, key_update, bdi_valid, done,
							bdi_eot, bdi_eoi, bdo_ready, decrypt, decrypt_r, bdi_eot_r, bdi_eoi_r,
							bdi_type, bdi_type_r, blockEmpty)
		BEGIN
		IF rst = '1' THEN
			hns <= IDLE;
		ELSE
			init <= '0';
			run <= '0';
			bdo_valid <= '0';
			bdi_ready <= '0';
			pn <= (OTHERS => '0');
			selV <= (OTHERS => '0');
			msg_auth_done <= '0';
			hns <= hps;
			key_ready <= '0';
			bdi_ready <= '0';
			npub_en <= '0';
			etag_en <= '0';
			tag_en <= '0';
			tag_ld <= '0';
			bdo_sel <= '0';
			bdi_eoi_l <= '0';
			bdi_eot_l <= '0';
			bdi_type_l <= '0';
			decrypt_l <= '0';
			
			CASE hps IS
				WHEN IDLE =>
					IF (bdi_valid = '1') THEN
						IF (key_update = '1') THEN
							hns <= ACTIVATE_KEY;
						ELSE
							hns <= LOAD_NPUB;
						END IF;
					END IF;
				WHEN ACTIVATE_KEY =>
					key_ready <= '1';
					IF (key_valid = '1') THEN
						hns <= LOAD_NPUB;
					END IF;
				WHEN LOAD_NPUB =>
					bdi_eoi_l <= '1';
					decrypt_l <= '1';
					IF (bdi_valid = '1') THEN
						bdi_ready <= '1';
						npub_en <= '1';
					END IF;
					IF (bdi_eot = '1') THEN
						hns <= LOAD_INIT;
					END IF;
				WHEN LOAD_INIT =>
					init <= '1';
					selV(1 DOWNTO 0) <= "01";
					hns <= RUN_INIT;
				WHEN RUN_INIT =>
					run <= '1';
					IF (done = '1') THEN
						hns <= LOAD_DATA;
					END IF;
				WHEN LOAD_DATA =>
					bdi_eot_l <= '1';
					bdi_eoi_l <= '1';
					bdi_type_l <= '1';
					decrypt_l <= '1';
					IF (bdi_eoi_r = '1') THEN
						bdi_ready <= '1';
						init <= '1';
						hns <= RUN_BLOCK;
					ELSIF (bdi_valid = '1') THEN
						bdi_ready <= '1';
						init <= '1';
						IF (bdi_type = "000") THEN
							IF (bdi_eot /= '1') THEN
								pn <= "11";
							END IF;
							IF (blockEmpty = '1') THEN
								hns <= LOAD_DATA;
							ELSE
								hns <= RUN_BLOCK;
							END IF;
						ELSIF (bdo_ready = '1') THEN
							IF (decrypt = '1') THEN
								selV(1 DOWNTO 0) <= "10"; --M/C || Vc
							END IF;
							hns <= RUN_BLOCK;
							bdo_valid <= '1';
							bdo_sel <= '1';
						END IF;
					END IF;
				WHEN RUN_BLOCK =>
					run <= '1';
					IF (done = '1') THEN
						hns <= LOAD_DATA;
						IF (bdi_type_r = "010" and bdi_eot_r = '1') THEN
							hns <= GENVER_TAG;
						END IF;
					END IF;
				WHEN GENVER_TAG =>
					IF (bdi_valid = '1' AND decrypt_r = '1') THEN
						bdi_ready <= '1';
						etag_en <= '1';
						IF (bdi_eot = '1') THEN --last etag block
							hns <= WRITE_TAG;
						END IF;
					ELSIF (bdo_ready = '1') THEN
						tag_ld <= '1';
						hns <= WRITE_TAG;
					END IF;
				when WRITE_TAG =>
					IF (decrypt_r = '1') THEN
						msg_auth_done <= '1';
						hns <= IDLE;
					ELSIF (bdo_ready = '1') THEN
						tag_en <= '1';
						bdo_valid <= '1';
					ELSE
						hns <= IDLE;
					END IF;
			END CASE;
		END IF;
	END PROCESS HANUMAN_OUTPUT;
END behavioral;